Baggins = AceLibrary("AceAddon-2.0"):new("AceConsole-2.0", "AceDB-2.0", "AceEvent-2.0", "AceDebug-2.0", "FuBarPlugin-2.0")
local pt = AceLibrary("PeriodicTable-2.0")
local L = AceLibrary("AceLocale-2.2"):new("Baggins")
local tablet = AceLibrary("Tablet-2.0")
local dewdrop = AceLibrary("Dewdrop-2.0")
local gratuity = AceLibrary("Gratuity-2.0")

Baggins.hasIcon = "Interface\\Icons\\INV_Jewelry_Ring_03"
Baggins.cannotDetachTooltip = true
Baggins.clickableTooltip = true
Baggins.independentProfile = true
Baggins.hideWithoutStandby = true

BINDING_HEADER_BAGGINS = L["Baggins"]
BINDING_NAME_BAGGINS_TOGGLEALL = L["Toggle All Bags"]

local itemstable = {}
local categorycache = {}
local useditems = {}
local slotcache = {}
local currentbag
local currentsection
local currentcategory

local BagNames = {
	[0] = "BackPack",
	[1] = "Bag1",
	[2] = "Bag2",
	[3] = "Bag3",
	[4] = "Bag4",
	[-2] = "KeyRing",
}

local BagTypes = { 
	--normal bags
	[0] = 1,
	[1] = 1,
	[2] = 1,
	[3] = 1,
	[4] = 1,
	--bank bags
	[-1] = 2,
	[5] = 2,
	[6] = 2,
	[7] = 2,
	[8] = 2,
	[9] = 2,
	[10] = 2,
	[11] = 2,
	--keyring
	[-2] = 3,		
}

local QualityNames = {
	[0] = "Poor",
	[1] = "Common",
	[2] = "Uncommon",
	[3] = "Rare",
	[4] = "Epic",
	[5] = "Legendary",
	[6] = "Artifact",
}

local ItemTypes = { 
	Armor = { "Cloth", "Idols", "Leather", "Librams", "Mail", "Miscellaneous", "Shields", "Totems", "Plate"},
	Consumable = { "Consumable" },
	Container  = {"Bag", "Enchanting Bag", "Engineering Bag", "Herb Bag", "Soul Bag" },
	Key = { "Key" },
	Miscellaneous = { "Junk" },
	Reagent  = { "Reagent" },
	Recipe  = { "Alchemy","Blacksmithing","Book","Cooking","Enchanting","Engineering","First Aid","Leatherworking","Tailoring" },
	Projectile  = {"Arrow","Bullet"},
	Quest  = { "Quest" },
	Quiver  = { "Ammo Pouch","Quiver" },
	["Trade Goods"]  = { "Devices","Explosives","Parts","Trade Goods" },
	Weapon  = { "Bows","Crossbows","Daggers","Guns","Fishing Pole","Fist Weapons","Miscellaneous","One-Handed Axes","One-Handed Maces","One-Handed Swords","Polearms","Staves","Thrown","Two-Handed Axes","Two-Handed Maces","Two-Handed Swords","Wands"},
}

local EquipLocs = {
	"INVTYPE_AMMO",
	"INVTYPE_HEAD",
	"INVTYPE_NECK",
	"INVTYPE_SHOULDER",
	"INVTYPE_BODY", 
	"INVTYPE_CHEST",
	"INVTYPE_ROBE",
	"INVTYPE_WAIST",
	"INVTYPE_LEGS",
	"INVTYPE_FEET",
	"INVTYPE_WRIST",
	"INVTYPE_HAND",
	"INVTYPE_FINGER",
	"INVTYPE_TRINKET",
	"INVTYPE_CLOAK",
	"INVTYPE_WEAPON",
	"INVTYPE_SHIELD",
	"INVTYPE_2HWEAPON",
	"INVTYPE_WEAPONMAINHAND",
	"INVTYPE_WEAPONOFFHAND",
	"INVTYPE_HOLDABLE",
	"INVTYPE_RANGED",
	"INVTYPE_THROWN",
	"INVTYPE_RANGEDRIGHT",
	"INVTYPE_RELIC",
	"INVTYPE_TABARD",
	"INVTYPE_BAG"
}
--removes any fields not used by the current rule type and sets up defaults if needed
function Baggins:CleanRule(rule)
	if rule.type ~= "ItemType" then
		rule.itype = nil
		rule.isubtype = nil
	else
		rule.itype = "Miscelaneous"
	end
	if rule.type ~= "ContainerType" then
		rule.ctype = nil
	else
		--rule.ctype = "
	end
	if rule.type ~= "Bag" then
		rule.bagid = nil
		rule.noempty = nil
	else
		rule.bagid = 0
	end
	if rule.type ~= "PTSet" then
		rule.setname = nil
	end
	if rule.type ~= "EquipLoc" then
		rule.equiploc = nil
	else
		rule.equiploc = EquipLocs[1]
	end
	if rule.type ~= "Quality" then
		rule.quality = nil
		rule.comp = nil
	else
		rule.quality = 1
		rule.comp = "=="
	end
end

local RuleTypes = {
	ItemType = {
		DisplayName = "Item Type",
		Description = "Filter by Item type and sub-type as returned by GetItemInfo",
		Matches = function(bag,slot,rule) 
			if not (rule.itype or rule.isubtype) then return end
			local link = GetContainerItemLink(bag, slot)
			if link then
				local _, _, _, _, _, Type, SubType = GetItemInfo(link)
				if Type and SubType then
					return Type == rule.itype and (rule.isubtype == nil or rule.isubtype == SubType)	
				end
			end
		end,
		GetName = function(rule) 
			return "ItemType - "..(rule.itype or "*")..":"..(rule.isubtype or "*")
		end,
		DewDropOptions = function(rule, level, value) 
			if level == 1 then
				dewdrop:AddLine('text', "Item Type Options", 'isTitle', true)
				dewdrop:AddLine('text', "Item Type", 'hasArrow', true,'value',"ItemType")
				dewdrop:AddLine('text', "Item Subtype", 'hasArrow', true,'value',"ItemSubtype")
				dewdrop:AddLine()

			elseif level == 2 and value == "ItemType" then	
				for k, v in pairs(ItemTypes) do
					dewdrop:AddLine('text', k, "checked", rule.itype == k,"func",function(k) rule.itype = k rule.isubtype = nil Baggins:OnRuleChanged() end,"arg1",k)
				end
			elseif level == 2 and value == "ItemSubtype" then
				dewdrop:AddLine('text', "All", "checked", rule.isubtype == nil,"func",function() rule.isubtype = nil Baggins:OnRuleChanged() end)
				dewdrop:AddLine()
				if rule.itype and ItemTypes[rule.itype] then
					for k, v in ipairs(ItemTypes[rule.itype]) do
						dewdrop:AddLine('text', v, "checked", rule.isubtype == v,"func",function(v) rule.isubtype = v Baggins:OnRuleChanged() end,"arg1",v)
					end
				end
			end
		end,
	},
	ContainerType = {
		DisplayName = "Container Type",
		Description = "Filter by the type of container the item is in.",	
		Matches = function(bag,slot,rule)
			if bag < 1 or bag > 4 then return end
			if not rule.ctype then return end
			local link = GetInventoryItemLink("player",19+bag)
			if link then
				local _, _, _, _, _, _, SubType = GetItemInfo(link)
				if SubType then
					return SubType == rule.ctype
				end
			end
		end,
		GetName = function(rule) 
			return "Container : "..(rule.ctype or "*none*")
		end,
		DewDropOptions = function(rule, level, value) 
			if level == 1 then
				dewdrop:AddLine('text', "Container Type Options", 'isTitle', true)
				dewdrop:AddLine('text', "Container Type", 'hasArrow', true,'value',"ContainerType")
				dewdrop:AddLine()		
			elseif level == 2 and value == "ContainerType" then
				dewdrop:AddLine('text', "All", "checked", rule.ctype == nil,"func",function() rule.ctype = nil Baggins:OnRuleChanged() end)
				dewdrop:AddLine()
				for k, v in ipairs(ItemTypes["Container"]) do
					dewdrop:AddLine('text', v, "checked", rule.ctype == v,"func",function(v) rule.ctype = v Baggins:OnRuleChanged() end,"arg1",v)
				end	
				for k, v in ipairs(ItemTypes["Quiver"]) do
					dewdrop:AddLine('text', v, "checked", rule.ctype == v,"func",function(v) rule.ctype = v Baggins:OnRuleChanged() end,"arg1",v)
				end		
			end
		end,
	},
	ItemID = {
		DisplayName = "ItemID",
		Description = "Filter by ItemID, this can be a space delimited list or ids to match.",			
		Matches = function(bag,slot,rule) 
			if not rule.ids then return end
			local link = GetContainerItemLink(bag, slot)
			if link then
				local itemid = string.match(link, "item:(%d+)")
				for match in string.gmatch(rule.ids,"(%d+)") do
					if itemid == match then 
						return true 
					end
				end		
			end
		end,
		GetName = function(rule) 
			if #(rule.ids or "") > 10 then
				return "ItemIDs "..(string.sub(rule.ids,1,10).."...")
			else
				return "ItemIDs "..(rule.ids or "*none*")
			end
		end,
		DewDropOptions = function(rule, level, value) 
			if level == 1 then
				dewdrop:AddLine('text', "ItemID Options", 'isTitle', true)
				dewdrop:AddLine('text', "Item IDs (space seperated list)", 'hasArrow', true,'value',"SetName","hasEditBox",true,"editBoxText", rule.ids or "","editBoxFunc", function(text) rule.ids = text Baggins:OnRuleChanged() end )
				dewdrop:AddLine()	
			end
		end,
	},
	Bag = {
		DisplayName = "Bag",
		Description = "Filter by the bag the item is in",	
		Matches = function(bag,slot,rule) 
			if rule.noempty then
				local link = GetContainerItemLink(bag, slot)
				if link then
					return bag == rule.bagid 
				end
			else
				return bag == rule.bagid
			end
		end,
		GetName = function(rule) 
			return "Bag "..(BagNames[rule.bagid] or rule.bagid or "*none*")..((rule.noempty and " *NotEmpty*") or "")
		end,
		DewDropOptions = function(rule, level, value) 
			if level == 1 then
				dewdrop:AddLine('text', "Bag Options", 'isTitle', true)
				dewdrop:AddLine('text', "Bag", 'hasArrow', true,'value',"Bag")
				dewdrop:AddLine('text', "Ignore Empty Slots", "checked", rule.noempty ,"func",function() rule.noempty = not rule.noempty Baggins:OnRuleChanged() end)
				dewdrop:AddLine()		
			elseif level == 2 and value == "Bag" then
				dewdrop:AddLine('text', "Backpack", "checked", rule.bagid == 0,"func",function() rule.bagid = 0 Baggins:OnRuleChanged() end)
				dewdrop:AddLine('text', "Bag1", "checked", rule.bagid == 1,"func",function() rule.bagid = 1 Baggins:OnRuleChanged() end)
				dewdrop:AddLine('text', "Bag2", "checked", rule.bagid == 2,"func",function() rule.bagid = 2 Baggins:OnRuleChanged() end)
				dewdrop:AddLine('text', "Bag3", "checked", rule.bagid == 3,"func",function() rule.bagid = 3 Baggins:OnRuleChanged() end)
				dewdrop:AddLine('text', "Bag4", "checked", rule.bagid == 4,"func",function() rule.bagid = 4 Baggins:OnRuleChanged() end)
				dewdrop:AddLine('text', "Key Ring", "checked", rule.bagid == -2,"func",function() rule.bagid = -2 Baggins:OnRuleChanged() end)
			end
		end,
	},
	PTSet = {
		DisplayName = "PeriodicTable Set",
		Description = "Filter by PeriodicTable Set",	
		Matches = function(bag,slot,rule) 
			if not rule.setname then return end
			local link = GetContainerItemLink(bag, slot)
			if link then
				local itemid = string.match(link, "item:(%d+)")
				itemid = tonumber(itemid)
				return pt:ItemInSet(itemid,rule.setname)
			end
		end,
		GetName = function(rule) 
			return "PTSet:"..(rule.setname or "*none*")
		end,
		DewDropOptions = function(rule, level, value) 
			if level == 1 then
				dewdrop:AddLine('text', "Periodic Table Set Options", 'isTitle', true)
				dewdrop:AddLine('text', "Set Name", 'hasArrow', true,'value',"SetName","hasEditBox",true,"editBoxText", rule.setname or "","editBoxFunc", function(text) rule.setname = text Baggins:OnRuleChanged() end )
				dewdrop:AddLine()		
			end
		end,
	},
	Empty = {
		DisplayName = "Empty Slots",
		Description = "Empty bag slots",
		Matches = function(bag,slot,rule) 
			if not (bag and slot) then return end
			if bag < 0 or bag > 4 then return end
			local link = GetContainerItemLink(bag, slot)
			if not link then 
				return true 
			end
		end,
		GetName = function(rule) 
			return "Empty Slots"
		end,
		DewDropOptions = function(rule, level, value) 
			if level == 1 then
				
			end
		end,
	},
	AmmoBag = {
		DisplayName = "Ammo Bag",
		Description = "Items in an ammo bag or quiver",
		Matches = function(bag,slot,rule) 
			if bag < 1 or bag > 4 then return end
			local link = GetInventoryItemLink("player",19+bag)
			if link then
				local _, _, _, _, _, Type = GetItemInfo(link)
				if Type then
					return Type == "Quiver"
				end
			end
		end,
		GetName = function(rule) 
			return "Ammo Bag Slots"
		end,
		DewDropOptions = function(rule, level, value)
			if level == 1 then
				
			end
		end,
	},
	Quality = {
		DisplayName = "Quality",
		Description = "Filter by Item Quality",
		Matches = function(bag,slot,rule) 
			if not (rule.comp and rule.quality) then return end
			local link = GetContainerItemLink(bag, slot)
			if link then
				local _, _, Rarity = GetItemInfo(link)
				if Rarity then
					return ( rule.comp == "==" and Rarity == rule.quality ) or
					       ( rule.comp == "<=" and Rarity <= rule.quality ) or
					       ( rule.comp == ">=" and Rarity >= rule.quality )
				end
			end
		end,
		GetName = function(rule) 
			return "Quality "..(rule.comp or "*none*").." "..(QualityNames[rule.quality] or rule.quality or "*none*")
		end,
		DewDropOptions = function(rule, level, value)
			if level == 1 then
				dewdrop:AddLine('text', "Quality Options", 'isTitle', true)
				dewdrop:AddLine('text', "Quality", 'hasArrow', true,'value',"Quality")
				dewdrop:AddLine('text', "Comparison", 'hasArrow', true, 'value', "QualComp")
				dewdrop:AddLine()			
			elseif level == 2 and value == "Quality" then
				dewdrop:AddLine('text', "Poor", "checked", rule.quality == 0,"func",function() rule.quality = 0 Baggins:OnRuleChanged() end)
				dewdrop:AddLine('text', "Common", "checked", rule.quality == 1,"func",function() rule.quality = 1 Baggins:OnRuleChanged() end)
				dewdrop:AddLine('text', "Uncommon", "checked", rule.quality == 2,"func",function() rule.quality = 2 Baggins:OnRuleChanged() end)
				dewdrop:AddLine('text', "Rare", "checked", rule.quality == 3,"func",function() rule.quality = 3 Baggins:OnRuleChanged() end)
				dewdrop:AddLine('text', "Epic", "checked", rule.quality == 4,"func",function() rule.quality = 4 Baggins:OnRuleChanged() end)
				dewdrop:AddLine('text', "Legendary", "checked", rule.quality == 5,"func",function() rule.quality = 5 Baggins:OnRuleChanged() end)
				dewdrop:AddLine('text', "Artifact", "checked", rule.quality == 6,"func",function() rule.quality = 6 Baggins:OnRuleChanged() end)
			elseif level == 2 and value == "QualComp" then		
				dewdrop:AddLine('text', "==", "checked", rule.comp == "==","func",function() rule.comp = "==" Baggins:OnRuleChanged() end)
				dewdrop:AddLine('text', "<=", "checked", rule.comp == "<=","func",function() rule.comp = "<=" Baggins:OnRuleChanged() end)
				dewdrop:AddLine('text', ">=", "checked", rule.comp == ">=","func",function() rule.comp = ">=" Baggins:OnRuleChanged() end)
			end
		end,
	},
	EquipLoc = {
		DisplayName = "Equip Location",
		Description = "Filter by Equip Location as returned by GetItemInfo",
		Matches = function(bag,slot,rule) 
			if not rule.equiploc then return end
			local link = GetContainerItemLink(bag, slot)
			if link then
				local _, _, _, _, _, _, _, _, EquipLoc = GetItemInfo(link)
				if EquipLoc then
					return EquipLoc == rule.equiploc
				end
			end
		end,
		GetName = function(rule) 
			return "Equip Location: "..(getglobal(rule.equiploc) or rule.equiploc or "*None*")
		end,
		DewDropOptions = function(rule, level, value) 
			if level == 1 then
				dewdrop:AddLine('text', "Equip Location Options", 'isTitle', true)
				dewdrop:AddLine('text', "Location", 'hasArrow', true,'value',"Location")
				dewdrop:AddLine()
			elseif level == 2 and value == "Location" then	
				for k, v in pairs(EquipLocs) do
					dewdrop:AddLine('text', getglobal(v) or v or "", "checked", rule.equiploc == v,"func",function(v) rule.equiploc = v Baggins:OnRuleChanged() end,"arg1",v)
				end
			end
		end,
	},
	Other = {
		DisplayName = "Unfiltered Items",
		Description = "Matches all items that arent matched by any other bag, NOTE: this should be the only rule in a category, others will be ignored",
		Matches = function(bag,slot,rule) 
			
			--local key = bag..":"..slot
			--return not useditems[key]
		end,
		GetName = function(rule) 
			return "Other"
		end,
		DewDropOptions = function(rule, level, value) 
			if level == 1 then
				
			end
		end,
	},
	Bind = {
		DisplayName = "Bind",
		Description = "Filter based on if the item binds, or if it is already bound",
		Matches = function(bag, slot, rule)
			if not rule.bindtype then return end
			if rule.bindtype == "Soulbound" then
				gratuity:SetBagItem(bag,slot)
				if gratuity:Find("Soulbound",2,2) then
					return true
				end
			elseif rule.bindtype == "Unbound" then
				local link = GetContainerItemLink(bag,slot)
					if link then
					gratuity:SetBagItem(bag,slot)
					if not gratuity:Find("Soulbound",2,2) then
						return true
					end
				end
			elseif rule.bindtype == "BoP" then
				local link = GetContainerItemLink(bag,slot)
				if link then
					gratuity:SetHyperlink(link)
					if gratuity:Find("Binds when picked up",2,2) then
						return true
					end
				end
			elseif rule.bindtype == "BoE" then
				local link = GetContainerItemLink(bag,slot)
				if link then
					gratuity:SetHyperlink(link)
					if gratuity:Find("Binds when equipped",2,2) then
						return true
					end
				end
			elseif rule.bindtype == "BoU" then
				local link = GetContainerItemLink(bag,slot,2,2)
				if link then
					gratuity:SetHyperlink(link)
					if gratuity:Find("Binds when used") then
						return true
					end
				end
			end
		end,
		GetName = function(rule)
			if rule.bindtype == "Soulbound" then
				return "Soulbound"
			elseif rule.bindtype == "Unbound" then
				return "Unbound"
			elseif rule.bindtype == "BoP" then
				return "Binds on pickup"
			elseif rule.bindtype == "BoE" then
				return "Binds on equip"
			elseif rule.bindtype == "BoU" then
				return "Binds on use"
			else
				return "Bind *unset*"
			end
		end,
		DewDropOptions = function(rule, level, value)
			if level == 1 then
				dewdrop:AddLine('text', "Bind Options", 'isTitle', true)
				dewdrop:AddLine('text', "Bind Type", 'hasArrow', true,'value',"BindType")
				dewdrop:AddLine()
			elseif level == 2 and value == "BindType" then	
				dewdrop:AddLine('text', "Soulbound", "checked", rule.bindtype == "Soulbound","func",function() rule.bindtype = "Soulbound" Baggins:OnRuleChanged() end)
				dewdrop:AddLine('text', "Unbound", "checked", rule.bindtype == "Unbound","func",function() rule.bindtype = "Unbound" Baggins:OnRuleChanged() end)
				dewdrop:AddLine('text', "Binds on pickup", "checked", rule.bindtype == "BoP","func",function() rule.bindtype = "BoP" Baggins:OnRuleChanged() end)
				dewdrop:AddLine('text', "Binds on equip", "checked", rule.bindtype == "BoE","func",function() rule.bindtype = "BoE" Baggins:OnRuleChanged() end)
				dewdrop:AddLine('text', "Binds on use", "checked", rule.bindtype == "BoU","func",function() rule.bindtype = "BoU" Baggins:OnRuleChanged() end)
			end
		end,
	}
}

--deep copy of a table, will NOT handle tables as keys or circular references
local function deepCopy(to, from)
	for k in pairs(to) do
		to[k] = nil
	end

	for k, v in pairs(from) do
		if type(v) == "table" then
			to[k] = {}
			deepCopy(to[k], from[k])
		else
			to[k] = from[k]
		end
	end
end
function Baggins:OnInitialize() 
    --self:RegisterEvent("BAG_UPDATE","OnBagUpdate")
    self.bagframes = {}
	self.itemcache = {}
	self.colors = {
		black = {r=0,g=0,b=0},
		white = {r=1,g=1,b=1}
	}
	
	self.opts = {
		type = "group",
		args = {
			[L["Columns"]] = {
				name = L["Columns"],
				type = "range",
				desc = L["Number of Columns shown in the bag frames"],
				order = 1,
				get = function() return self.db.profile.columns end,
				set = function(value) self.db.profile.columns = value self:UpdateBags() end,
				min = 2,
				max = 20,
				step = 1,
			},
			[L["Layout"]] = {
				name = L["Layout"],
				type = "group",
				desc = L["Layout of the bag frames."],
				order = 2,
				args = {
					[L["Automatic"]] = {
						name = L["Automatic"],
						type = "toggle",
						desc = L["Automatically arrange the bag frames as the default ui does"],
						get = function() return self.db.profile.layout == "auto" end,
						set = function() self.db.profile.layout = "auto" end,
					},
					[L["Manual"]] = {
						name = L["Manual"],
						type = "toggle",
						desc = L["Each bag frame can be positioned manually."],
						get = function() return self.db.profile.layout == "manual" end,
						set = function() self.db.profile.layout = "manual" end,
					},
				},
			},
			[L["Show Section Title"]] = {
				name = L["Show Section Title"],
				type = "toggle",
				desc = L["Show a title on each section of the bags"],
				order = 3,
				get = function() return self.db.profile.showsectiontitle end,
				set = function(value) self.db.profile.showsectiontitle = value 	
							self:UpdateBags()
						end
			},		
			[L["Hide Empty Sections"]] = {
				name = L["Hide Empty Sections"],
				type = "toggle",
				desc = L["Hide sections that have no items in them."],
				order = 4,
				get = function() return self.db.profile.hideemptysections end,
				set = function(value) self.db.profile.hideemptysections = value 	
							self:UpdateBags()
						end
			},		
			CompressEmptySlots = {
				name = "Compress Empty Slots",
				type = "toggle",
				desc = "Show all empty slots as a single button with a count on it",
				order = 5,
				get = function() return self.db.profile.compressempty end,
				set = function(value) self.db.profile.compressempty = value 	
							self:UpdateBags()
						end,
			},	
			[L["Sort"]] = {
				name = L["Sort"],
				type = "group",
				desc = L["How items are sorted"],
				order = 5,
				args = {
					[L["Quality"]] = {
						name = L["Quality"],
						type = "toggle",
						desc = L["Items are sorted by quality."],
						get = function() return self.db.profile.sort == "quality" end,
						set = function() self.db.profile.sort = "quality" end,
					},
					[L["Name"]] = {
						name = L["Name"],
						type = "toggle",
						desc = L["Items are sorted by name."],
						get = function() return self.db.profile.sort == "name" end,
						set = function() self.db.profile.sort = "name" end,
					},
				},
			},		
			spacer1 = {
				type = "header",
				order = 49,
			},				
			EditBags = {
				name = "Edit Bags",
				type = "execute",
				order = 50,
				desc = "Edit the Bag Definitions",
				func = function() self:OpenBagEditTablet() end,
			},
			EditCategories = {
				name = "Edit Categories",
				type = "execute",
				order = 51,
				desc = "Edit the Category Definitions",
				func = function() self:OpenCategoryEditTablet() end,
			},
			spacer2 = {
				type = "header",
				order = 99,
			},			
			LoadProfile = {
				name = "Load Profile",
				type = "group",
				desc = "Load a built-in profile: NOTE: ALL Custom Bags will be lost and any edited built in categories will be lost.",
				order = 100,
				args = {
					Default = {
						name = "Default",
						type = "execute",
						desc = "A default set of bags sorting your inventory into categories",
						func = function() self:ApplyProfile(self.profiles.default)	end,
					},
					AllInOne = {
						name = "All in one",
						type = "execute",
						desc = "A single bag containing your whole inventory, sorted by quality",
						func = function() self:ApplyProfile(self.profiles.allinone)	end,
					},
				},
			},				
	
		}			
	}
	
	self.OnMenuRequest = self.opts
	
	self:RegisterDB("BagginsDB")
	self:RegisterDefaults('profile', {
		showsectiontitle = true,
		hideemptysections = true,
		compressempty = true,
		columns = 5,
		sorttype = "quality",
		layout = "auto",
		bags = { },
		categories = {}
	}
	)
	
	self.profiles = {
		allinone = {
			layout = "manual",
			columns = 12,
			sorttype = "quality",
			showsectiontitle = true,
			bags = { 
				{ 
					name = "All In One",
					openWithAll = true,
					sections = { { name="Bags", cats={"Bags"} }, { name="Ammo Bag", cats={"AmmoBag"} }, { name="Keyring", cats={"Keyring"} } } 
				}
			}
		},
		default = {
			showsectiontitle = true,
			columns = 5,
			sorttype = "quality",
			layout = "auto",
			bags = {
				{
					name = "Consumables",
					openWithAll = true,
					sections = 
					{
						{cats = {"Water"}, name = "Water",}, 
						{cats = {"Food"}, name = "Food",},
						{cats = {"FirstAid"}, name = "First Aid",},
						{cats = {"Potions"}, name = "Potions",},
						{cats = {"Scrolls"}, name = "Scrolls",},
						{cats = { "Misc Consumables" }, name = "Misc"},
					}
				},
				{
					name = "Equipment",
					openWithAll = true,
					sections = { { name="Armor", cats={ "Armor" } }, { name="Weapons", cats={ "Weapons" } } }
				},
				{
					name = "Quest",
					openWithAll = true,
					sections = { { name="Quest Items", cats = { "Quest" } } }
				},	
				{
					name = "Ammo",
					openWithAll = true,
					sections = { { name="Ammo", cats={ "AmmoBag" } }, { name="SoulShards", cats={ "SoulBag" } } }
				},		
				{
					name = "Other",
					openWithAll = true,
					sections = { { name="Other", cats = { "Other" } }, { name="Trash", cats = { "Trash", "TrashEquip" } }, { name="Empty", cats = { "Empty" } } }
				}
			}
		}
	}
	
	self.defaultcategories = {
		["Misc Consumables"] = { name="Misc Consumables", {type="ItemType" ,itype = "Consumable"},{operation = "NOT",type = "PTSet",setname = "All Edible Food",},{operation = "NOT",type = "PTSet",setname = "Water",},{operation = "NOT",type = "PTSet",setname = "Potion",},{operation = "NOT",type = "PTSet",setname = "Scrolls",},{operation = "NOT",type = "PTSet",setname = "Bandages",},},
		Armor = { name="Armor", { type="ItemType", itype="Armor" } },
		Weapons = { name="Weapons", { type="ItemType", itype="Weapon" }, { type="PTSet", setname="Tradeskill Tools", operation="NOT" }, },
		Quest = { name="Quest", { type="ItemType", itype="Quest" } },
		AmmoBag = { name="AmmoBag", { type="AmmoBag" } },
		Trash = { name="Trash", { type="Quality", quality = 0, comp = "<=" } },
		TrashEquip = { name="TrashEquip", { type="ItemType", itype="Armor" }, { type="ItemType", itype="Weapon", operation="OR" }, { type="Quality", quality = 1, comp = "<=", operation="AND" }, { type="PTSet", setname="Tradeskill Tools", operation="NOT" }, },
		Other = { name="Other", { type="Other" } },
		Empty = { name="Empty", { type="Empty" }, { type="AmmoBag", operation="NOT"}, { type="ContainerType", ctype="Soul Bag", operation="NOT"}, },
		Bags = { name="Bags", { type="Bag", bagid=1 }, { type="Bag", bagid=2, operation="OR" }, { type="Bag", bagid=3, operation="OR" }, { type="Bag", bagid=4, operation="OR" }, { type="Bag", bagid=0, operation="OR" }, { type="AmmoBag", operation="NOT"} },
		Keyring = { name="Keyring", { type="Bag", noempty=true, bagid=-2 }, },
		Potions = { name="Potions", { type="PTSet", setname="Potion" }, },
		Food = { name="Food", { type="PTSet", setname="All Edible Food" }, },
		Scrolls = { name="Scrolls", { type="PTSet", setname="Scrolls" }, },
		FirstAid = { name="FirstAid", { type="PTSet", setname="CraftedBy - First Aid" }, },
		Water = { name="Water", { type="PTSet", setname="Water" }, },
		["AQ Quest"] = { name="AQ Quest", { type="PTSet", setname="AQ - Quest Items" }, },
		["ZG Quest"] = { name="ZG Quest", { type="PTSet", setname="ZG - Quest Items" }, },
		SoulBag = { name="SoulBag", { type="ContainerType", ctype="Soul Bag" }, },
	}

	for k, v in ipairs(self.db.profile.bags) do
		self:CreateBagFrame(k)
	end
	self:RegisterEditTablets()
end

function Baggins:OnEnable()
	self:RegisterEvent("BAG_UPDATE","OnBagUpdate")
	self:RegisterEvent("BAG_UPDATE_COOLDOWN", "UpdateItemButtons");
	self:RegisterEvent("ITEM_LOCK_CHANGED", "UpdateItemButtons");
	--if this is a first time load, fill the inital categories and bags
	--cant leave these in the defaults since removing a bag would have it come back on reload
	local catsexist
	for k in pairs(self.db.profile.categories) do
		catsexist = true
		break
	end
	if not catsexist then
		deepCopy(self.db.profile.categories, self.defaultcategories)
	end
	 if #self.db.profile.bags == 0 then
		deepCopy(self.db.profile.bags, self.profiles.default.bags)
	end
	--force an update of all bags on first opening
	self.doInitialUpdate = true
end

function Baggins:OnDisable()
	self:HideAllBags()
end

function Baggins:OnBagUpdate(bagid)
	--ignore bank bag updates
	if bagid > 4 or bagid == -1 then return end
	self:CheckSlotsChanged(bagid)
	self:UpdateBags()
end

function Baggins:ForceFullUpdate()
	local bagid
	for bagid = 0, 4 do
		self:CheckSlotsChanged(bagid, true)
	end
	self:CheckSlotsChanged(-2,true)
end

function Baggins:CountEmptySlots()
	local bag, slot
	local count = 0
	for bag = 0, 4 do
		for slot = 1, GetContainerNumSlots(bag) do
			if not GetContainerItemLink(bag,slot) then
				count = count + 1
			end
		end
	end
	return count
end

function Baggins:CheckSlotsChanged(bag, forceupdate)
	local slot
	for slot = 1, GetContainerNumSlots(bag) do
		local key = bag..":"..slot
		local iteminfo
		
		local link = GetContainerItemLink(bag, slot)
		local _, itemCount = GetContainerItemInfo(bag, slot)
		

		if link then
			local itemid = string.match(link, "item:(%d+)")
			local itemName = GetItemInfo(link)
			iteminfo = itemid.." "..itemCount.." "..(itemName or "_")
		end
		
		if slotcache[key] ~= iteminfo or forceupdate then
			self:OnSlotChanged(bag, slot)
			slotcache[key] = iteminfo
		end
	end
end

function Baggins:OnSlotChanged(bag, slot)
	
	local key = bag..":"..slot
	if not useditems[key] then
		useditems[key] = false
	end
	for catid, category in pairs(self.db.profile.categories) do
		if not categorycache[catid] then
			categorycache[catid] = {}
		end
		categorycache[catid][key] = nil

		local catmatch = false
		categorycache[catid]["Other"] = nil
		for ruleid, rule in ipairs(category) do
			local rulematch
			if rule.type == "Other" then
				categorycache[catid]["Other"] = true
			else
				if ruleid == 1 then 
					operation = "OR"
				else
					operation = rule.operation or "OR"
				end
				rulematch = (RuleTypes[rule.type] and RuleTypes[rule.type].Matches(bag,slot,rule))
				if operation == "OR" then
					catmatch = catmatch or rulematch
				elseif operation == "AND" then
					catmatch = catmatch and rulematch
				elseif operation == "NOT" then
					catmatch = catmatch and (not rulematch)
				end
			end
		end
		if catmatch then
			categorycache[catid][key] = true
		end
	end
end

function Baggins:ApplyProfile(profile)
	self:CloseAllBags()
	for k, v in pairs(profile) do  
		if type(v) == "table" then
			if not self.db.profile[k] then
				self.db.profile[k] = {}
			end
			deepCopy(self.db.profile[k],v)
		else
			self.db.profile[k] = v
		end
	end 
	for k, v in pairs(self.defaultcategories) do
		if not self.db.profile.categories[k] then
			self.db.profile.categories[k] = {}
		end
		deepCopy(self.db.profile.categories[k], v)
	end
	tablet:Refresh("BagginsEditCategories")
	tablet:Refresh("BagginsEditBags")
	self:ForceFullUpdate()
	self:UpdateBags() 
	self:OpenAllBags()
end

local othersections = {}

function Baggins:UpdateBags()
	self:StartBags()
	local isVisible = false
	for bagid, bag in pairs(self.db.profile.bags) do
		if self.bagframes[bagid] and self.bagframes[bagid]:IsVisible() then
			isVisible = true
		end
	end
	if not isVisible then return end
	for bagid, bag in pairs(self.db.profile.bags) do
		if self.bagframes[bagid] then
			self:StartBag(self.bagframes[bagid])
			self:SetBagTitle(bagid,bag.name)
			for sectionid, section in pairs(bag.sections) do
				self:StartSection()
				for catid, catname in pairs(section.cats) do
					if not categorycache[catname]["Other"] then
						self:AppendCategory(catname)
					else
						othersections[bagid] = sectionid
					end
				end
				if self.bagframes[bagid]:IsVisible() then
					self:SetSection(bagid,sectionid,section.name,self:FinishSection())
				end
			end
			self:UpdateBagFrameSize(bagid)
		end
	end
	for bagid, sectionid in pairs(othersections) do
		self:StartSection()
		self:AppendOther()
		self:SetSection(bagid,sectionid,self.db.profile.bags[bagid].sections[sectionid].name,self:FinishSection())
		self:UpdateBagFrameSize(bagid)
		othersections[bagid] = nil
	end
end

function Baggins:NewBag(bagname)
	tinsert(self.db.profile.bags, { name=bagname, openWithAll=true, sections={}})
	if not self.bagframes[#self.db.profile.bags] then
		self:CreateBagFrame(#self.db.profile.bags)
	end
	currentbag = #self.db.profile.bags
	self:UpdateBags()
	if self.db.profile.layout == "auto" then
		self:LayoutBagFrames()
	end
end

function Baggins:RemoveBag(bagid)
	self:CloseBag(bagid)
	tremove(self.db.profile.bags, bagid)
	self:UpdateBags()
	if self.db.profile.layout == "auto" then
		self:LayoutBagFrames()
	end	
end

function Baggins:NewSection(bagid,sectionname)
	tinsert(self.db.profile.bags[bagid].sections, { name=sectionname, cats = {} })
	currentsection = #self.db.profile.bags[bagid].sections
	self:UpdateBags()
	if self.db.profile.layout == "auto" then
		self:LayoutBagFrames()
	end	
end

function Baggins:NewCategory(catname)
	if not self.db.profile.categories[catname] then
		self.db.profile.categories[catname] = { name=catname } 
		currentcategory = catname
	else
		StaticPopup_Show("BAGGINS_CATEXISTS")
	end
end

function Baggins:RemoveSection(bagid, sectionid)
	tremove(self.db.profile.bags[bagid].sections, sectionid)
	self:UpdateBags()
	if self.db.profile.layout == "auto" then
		self:LayoutBagFrames()
	end	
end

function Baggins:RemoveRule(catid, ruleid)
	tremove(self.db.profile.categories[catid], ruleid)
	tablet:Refresh("BagginsEditCategories") 
	self:ForceFullUpdate()
end

function Baggins:AddCategory(bagid,sectionid,category)
	tinsert(self.db.profile.bags[bagid].sections[sectionid].cats,category)
	self:UpdateBags()
	if self.db.profile.layout == "auto" then
		self:LayoutBagFrames()
	end	
end

function Baggins:RemoveCategory(bagid,sectionid,catid)
	if type(bagid) == "string" then
		--removing a category completely
		local inUse = false
		for bid, bag in ipairs(self.db.profile.bags) do
			for sid, section in ipairs(bag.sections) do
				for cid, cat in ipairs(section.cats) do
					if cat == bagid then
						inUse = true
					end
				end
			end
		end
		if not inUse then
			self.db.profile.categories[bagid] = nil
		else
			StaticPopup_Show("BAGGINS_CATINUSE")
		end	
	else
		--removing a category from a bag
		tremove(self.db.profile.bags[bagid].sections[sectionid].cats,catid)
		self:UpdateBags()
		if self.db.profile.layout == "auto" then
			self:LayoutBagFrames()	
		end	
	end
end

function Baggins:SaveBagPosition(bagid)
	local frame = self.bagframes[bagid]
	if not frame then return end
	local s = frame:GetEffectiveScale()
	self.db.profile.bags[bagid].x = frame:GetLeft() * s
	self.db.profile.bags[bagid].y = frame:GetTop() * s
end

function Baggins:LoadBagPosition(bagid)		
	local frame = self.bagframes[bagid]
	if not frame then return end
	local s = frame:GetEffectiveScale()
	frame:ClearAllPoints()
	if self.db.profile.bags[bagid].x and self.db.profile.bags[bagid].y then
		frame:SetPoint("TOPLEFT",UIParent,"BOTTOMLEFT", self.db.profile.bags[bagid].x / s, self.db.profile.bags[bagid].y / s)
	else
		frame:SetPoint("CENTER",UIParent,"CENTER",0,0)
	end
end

function Baggins:EmptyItemsTable()
	for k, v in pairs(itemstable) do
		itemstable[k] = nil
	end
end

---------------------------------
-- Category Building Functions
----------------------------------

function Baggins:StartBags()
	for k, v in pairs(useditems) do
		useditems[k] = false
	end
end

--start filling a bag, sets all sections to hidded untill they are filled
function Baggins:StartBag(bag)
	for sectionid, sectionframe in ipairs(bag.sections) do
		sectionframe.used = false
		sectionframe:Hide()
		for itemid,itemframe in ipairs(sectionframe.items) do
			itemframe:Hide()
		end
	end
end

--starting a new section, clear the itemstable
function Baggins:StartSection()
	self:EmptyItemsTable()
end

--finished a section, return the item list
function Baggins:FinishSection()
	return itemstable
end

--appends a category to the itemslist
function Baggins:AppendCategory(catname)
	if not categorycache[catname] then return end
	if not categorycache[catname]["Other"] then
		for k, v in pairs(categorycache[catname]) do
			if itemstable[k] == nil then
				tinsert(itemstable, k)
				itemstable[k] = true
				useditems[k] = true
			end
		end
	end
end

function Baggins:AppendOther()
	for k, v in pairs(useditems) do
		if not v then
			local bag = string.match(k, "(-?%d+):")
			if bag then
				bag = tonumber(bag)
				if bag >= 0 and bag <= 4 then
					tinsert(itemstable, k)
					itemstable[k] = true
				end
			end
		end
	end
end


function Baggins:PTCategoryToTable(base,cattable)
	for k,v in ipairs(cattable) do
		if type(base[v]) == "table" then
			cattable[v] = self:PTCategoryToTable(base,base[v])
			cattable[k] = nil
			base[v] = false
		else
			cattable[v] = true
			cattable[k] = nil
			base[v] = false
		end
	end
	return cattable
end

local BagFrameBackdrop = {
	bgFile = "Interface/Tooltips/UI-Tooltip-Background", 
	edgeFile = "Interface/Tooltips/UI-Tooltip-Border", 
	tile = true, tileSize = 16, edgeSize = 16, 
	insets = { left = 5, right = 5, top = 5, bottom = 5 }
}
	
function Baggins:CreateBagFrame(bagid)
	if not bagid then return end
	local bagname = "BagginsBag"..bagid
	
	local frame = CreateFrame("Frame",bagname,UIParent)
	self.bagframes[bagid] = frame
	frame.bagid = bagid
	frame:SetWidth(100)
	frame:SetHeight(100)
	frame:SetPoint("CENTER",UIParent,"CENTER",0,0)
	frame:EnableMouse(true)
	frame:SetMovable(true)
	frame:SetScript("OnMouseDown",function() this:StartMoving() end)
	frame:SetScript("OnMouseUp",function() this:StopMovingOrSizing() self:SaveBagPosition(this.bagid) end)
	frame:SetBackdrop(BagFrameBackdrop)
	frame:SetBackdropColor(self.colors.Black)
	
	frame.closebutton = CreateFrame("Button",bagname.."CloseButton",frame,"UIPanelCloseButton")
	frame.closebutton:SetScript("OnClick", function() self:CloseBag(this:GetParent().bagid) end)
	frame.closebutton:SetPoint("TOPRIGHT",frame,"TOPRIGHT",0,0)
	
	frame.title = frame:CreateFontString(bagname.."Title","OVERLAY","GameFontNormalLarge")
	frame.title:SetText("Baggins")
	frame.title:SetVertexColor(1,1,1,1)
	frame.title:SetPoint("TOPLEFT",frame,"TOPLEFT",10,-10)
	frame.title:SetHeight(12)
	frame:SetFrameStrata("HIGH")
	frame:SetClampedToScreen(true) 
	frame.sections = {}
	frame.bagparents = {}

	self:UpdateBagFrameSize(bagid)
	frame:Hide()
end

function Baggins:SetBagTitle(bagid,title)
	if self.bagframes[bagid] then
		self.bagframes[bagid].title:SetText(title)
		self:UpdateBagFrameSize(bagid)
	end
end

function Baggins:SetSection(bagid, secid,title, contents)
	local bagframe = self.bagframes[bagid]
	if not bagframe then return end
	local i
	for i = 1,secid do
		if not bagframe.sections[i] then
			bagframe.sections[i] = self:CreateSectionFrame(bagframe,i)
			if i == 1 then
				bagframe.sections[i]:SetPoint("TOPLEFT",bagframe,"TOPLEFT",10,-36)
			else
				bagframe.sections[i]:SetPoint("TOPLEFT",bagframe.sections[i-1],"BOTTOMLEFT",0,1)
			end
		end
	end
	self:UpdateSectionContents(bagframe.sections[secid],title,contents)
	self:UpdateBagFrameSize(bagid)
end

function Baggins:UpdateBagFrameSize(bagid)
	local bagframe = self.bagframes[bagid]
	if not bagframe then return end
	--10px left offset + 32px close button
	local width = bagframe.title:GetWidth() + 42
	--32px heading
	local height = 37
	for id,section in pairs(bagframe.sections) do
		if section.used then
			width = math.max(width,section:GetWidth()+20)
			height = height + section:GetHeight()
		end
	end
	bagframe:SetWidth(width)
	bagframe:SetHeight(height)
end

function Baggins:CreateSectionFrame(bagframe,secid)
	local frame = CreateFrame("Frame",bagframe:GetName().."Section"..secid,bagframe)
	
	frame.title = frame:CreateFontString(bagframe:GetName().."SectionTitle","OVERLAY","GameFontNormalSmall")
	frame.title:SetVertexColor(1,1,1,1)
	frame.title:SetText("")
	frame.title:SetPoint("TOPLEFT",frame,"TOPLEFT",0,0)
	frame.title:SetHeight(13)
	frame:SetFrameStrata("HIGH")
	frame:Show()
	frame.items = {}
	return frame
end

function Baggins:UpdateSectionContents(sectionframe,title,itemlist)

	local totalwidth, totalheight = 1,1
	local cols = self.db.profile.columns
	local buttonnum = 1
	local usedEmpty = false
	local isEmpty
	if self.db.profile.showsectiontitle then
		sectionframe.title:SetText(title)
		totalheight = 23
	else
		sectionframe.title:SetText("")
		totalheight = 10
	end
	
	self:SortItemList(itemlist,self.db.profile.sorttype)
	local itemcount = 0
	for k, item in ipairs(itemlist) do
		itemcount = itemcount + 1
		if itemlist[item] then
			bag, slot = string.match(item,"^(-?%d+):(%d+)$")
			if not slotcache[item] then
				isEmpty = true
			end
			if bag and slot then
				bag = tonumber(bag)
				slot = tonumber(slot)
				
				if not isEmpty or (isEmpty and not usedEmpty) then

					sectionframe.items[buttonnum] = sectionframe.items[buttonnum] or self:CreateItemButton(sectionframe,buttonnum)
					local itemframe = sectionframe.items[buttonnum]
					if isEmpty and self.db.profile.compressempty then 
						usedEmpty = true
						itemframe.countoverride = self:CountEmptySlots()
					else
						itemframe.countoverride = nil
					end
					self:UpdateItemButton(sectionframe:GetParent(),itemframe,bag,slot)
					
					if buttonnum == 1 then
						if self.db.profile.showsectiontitle then
							itemframe:SetPoint("TOPLEFT",sectionframe.title,"BOTTOMLEFT",0,0)
						else
							itemframe:SetPoint("TOPLEFT",sectionframe,"TOPLEFT",0,0)
						end
						totalwidth = 37
						totalheight = totalheight+39
					elseif buttonnum%cols == 1 then
						itemframe:SetPoint("TOPLEFT",sectionframe.items[buttonnum-cols],"BOTTOMLEFT",0,-2)
						totalheight = totalheight+39
					else
						itemframe:SetPoint("TOPLEFT",sectionframe.items[buttonnum-1],"TOPRIGHT",2,0)
						if buttonnum <= cols then
							totalwidth = totalwidth+39
						end
					end
					itemframe:Show()
					buttonnum = buttonnum + 1
				end
			end
		end
	end
	
	sectionframe.used = true
	sectionframe:Show()
	--hide any extra buttons
	while sectionframe.items[buttonnum] do
		sectionframe.items[buttonnum]:Hide()
		buttonnum = buttonnum + 1
	end
	if self.db.profile.showtitle then
		totalwidth = max(totalwidth,sectionframe.title:GetWidth())
	end
	
	if itemcount == 0 and self.db.profile.hideemptysections then
		totalheight = 1
		sectionframe:Hide()
	end
	sectionframe:SetWidth(totalwidth)
	sectionframe:SetHeight(totalheight)	
end

local function NameComp(a, b)
	local baga, slota = string.match(a,"^(-?%d+):(%d+)$")
	local bagb, slotb = string.match(b,"^(-?%d+):(%d+)$")
	local linka = GetContainerItemLink(baga, slota)
	local linkb = GetContainerItemLink(bagb, slotb)
	if not linka then
		return false
	end
	if not linkb then
		return true
	end
	local namea = GetItemInfo(linka)
	local nameb = GetItemInfo(linkb)
	if not namea then return true end
	if not nameb then return false end		
	return namea < nameb
end

local function QualityComp(a, b)
	local baga, slota = string.match(a,"^(-?%d+):(%d+)$")
	local bagb, slotb = string.match(b,"^(-?%d+):(%d+)$")
	local linka = GetContainerItemLink(baga, slota)
	local linkb = GetContainerItemLink(bagb, slotb)
	if not linka then
		return false
	end
	if not linkb then
		return true
	end

	local namea, _, quala = GetItemInfo(linka)
	local nameb, _, qualb = GetItemInfo(linkb)
	if not quala then return true end
	if not qualb then return false end
	if quala == qualb then
		return namea < nameb
	else
		return quala < qualb
	end
end

function Baggins:SortItemList(itemlist, sorttype)
	local func
	if sorttype == "name" then
		func = NameComp
	elseif sorttype == "quality" then
		func = QualityComp 
	else
		return
	end
	table.sort(itemlist,func)
end


function Baggins:CreateItemButton(sectionframe,item)
	local frame = CreateFrame("Button",sectionframe:GetName().."Item"..item,frame,"ContainerFrameItemButtonTemplate",sectionframe)
	frame:ClearAllPoints()
	local cooldown = getglobal(frame:GetName().."Cooldown")
	cooldown:SetAllPoints(frame)
	cooldown:SetFrameLevel(10)
	frame:SetFrameStrata("HIGH")
	frame:Show()
	return frame
end

function Baggins:CloseBag(bagid)
	if self.bagframes[bagid] then
		self.bagframes[bagid]:Hide()
		if self.db.profile.layout == "auto" then
			self:LayoutBagFrames()
		end
	end
end

function Baggins:CloseAllBags()
	for bagid, bag in ipairs(self.db.profile.bags) do
		Baggins:CloseBag(bagid)	
	end
end

function Baggins:ToggleBag(bagid)
	if not self.bagframes[bagid] then
		self:CreateBagFrame(bagid)
		self:OpenBag(bagid)
	elseif self.bagframes[bagid]:IsVisible() then
		self:CloseBag(bagid)
	else
		self:OpenBag(bagid)
	end
end
function Baggins:OpenBag(bagid,noupdate)
	if self.doInitialUpdate then
		self:ForceFullUpdate()
		self.doInitialUpdate = nil
	end
	if not self.bagframes[bagid] then
		self:CreateBagFrame(bagid)
	end
	self.bagframes[bagid]:Show()
	if not noupdate then
		self:UpdateBags()
	end
	if self.db.profile.layout == "auto" then
		self:LayoutBagFrames()
	else
		self:LoadBagPosition(bagid)
	end

end

-- "All Bags" in these 3 functions refers to bags that are set to openWithAll
function Baggins:OpenAllBags()
	for bagid, bag in ipairs(self.db.profile.bags) do
		if bag.openWithAll then
			Baggins:OpenBag(bagid,true)
		end			
	end
	self:UpdateBags()
	if self.db.profile.layout == "auto" then
		self:LayoutBagFrames()
	else
		self:LoadBagPosition(bagid)
	end	
end

function Baggins:ToggleAllBags()
	if self:IsAllOpen() then
		self:CloseAllBags()
	else
		self:OpenAllBags()
	end
end

function Baggins:IsAllOpen()
	for bagid, bag in ipairs(self.db.profile.bags) do
		if bag.openWithAll then
			if not ( self.bagframes[bagid] and self.bagframes[bagid]:IsVisible() ) then
				return false
			end
		end			
	end
	return true
end

function Baggins:UpdateItemButtons()
	for bagid, bag in ipairs(self.bagframes) do
		for sectionid, section in ipairs(bag.sections) do
			for buttonid, button in ipairs(section.items) do
				self:UpdateItemButton(bag,button,button:GetParent():GetID(),button:GetID())
			end
		end
	end
end
function Baggins:UpdateItemButton(bagframe,button,bag,slot)
	local texture, itemCount, locked, quality, readable = GetContainerItemInfo(bag, slot);
	button:SetID(slot)
	if not bagframe.bagparents[bag] then 
		bagframe.bagparents[bag] = CreateFrame("Frame",nil,bagframe)
		bagframe.bagparents[bag]:SetID(bag)
	end
	button:SetParent(bagframe.bagparents[bag])
	SetItemButtonTexture(button, texture)
	if button.countoverride then
		SetItemButtonCount(button, button.countoverride)
	else
		SetItemButtonCount(button, itemCount)
	end
	SetItemButtonDesaturated(button, locked, 0.5, 0.5, 0.5)
	
	if ( texture ) then
		self:UpdateItemButtonCooldown(bag, button);
		button.hasItem = 1;
	else
		getglobal(button:GetName().."Cooldown"):Hide()
		button.hasItem = nil;
	end

	button.readable = readable;
	--local normalTexture = getglobal(name.."Item"..j.."NormalTexture");
	--if ( quality and quality ~= -1) then
	--	local color = getglobal("ITEM_QUALITY".. quality .."_COLOR");
	--	normalTexture:SetVertexColor(color.r, color.g, color.b);
	--else
	--	normalTexture:SetVertexColor(TOOLTIP_DEFAULT_COLOR.r, TOOLTIP_DEFAULT_COLOR.g, TOOLTIP_DEFAULT_COLOR.b);
	--end
	local showSell = nil;
	if ( GameTooltip:IsOwned(button) ) then
		if ( texture ) then
			local hasCooldown, repairCost = GameTooltip:SetBagItem(bag,slot);
			if ( hasCooldown ) then
				button.updateTooltip = TOOLTIP_UPDATE_TIME;
			else
				button.updateTooltip = nil;
			end
			
			if ( InRepairMode() and (repairCost > 0) ) then
				GameTooltip:AddLine(TEXT(REPAIR_COST), "", 1, 1, 1);
				SetTooltipMoney(GameTooltip, repairCost);
				GameTooltip:Show();
			elseif ( MerchantFrame:IsShown() and not locked) then
				showSell = 1;
			end
		else
			GameTooltip:Hide();
		end
		if ( showSell ) then
			ShowContainerSellCursor(bag, slot);
		elseif ( readable ) then
			ShowInspectCursor();
		else
			ResetCursor();
		end
	end
end

function Baggins:UpdateItemButtonCooldown(container, button)
	local cooldown = getglobal(button:GetName().."Cooldown")
	local start, duration, enable = GetContainerItemCooldown(container, button:GetID())
	CooldownFrame_SetTimer(cooldown, start, duration, enable)
	if ( duration > 0 and enable == 0 ) then
		SetItemButtonTextureVertexColor(button, 0.4, 0.4, 0.4);
	end
end

function Baggins:LayoutBagFrames()
	local frame, xOffset, yOffset, screenHeight, freeScreenHeight, column, maxwidth
	local screenWidth = GetScreenWidth()

	screenHeight = GetScreenHeight()
	
	-- Adjust the start anchor for bags depending on the multibars
	xOffset = 36
	yOffset = 50
	-- freeScreenHeight determines when to start a new column of bags
	freeScreenHeight = screenHeight - yOffset - 200;
	maxwidth = 1
	for index, frame in ipairs(self.bagframes) do
		maxwidth = math.max(maxwidth, frame:GetWidth())
	end
	column = 0
	local index = 1
	local prevframe
	for id, frame in ipairs(self.bagframes) do
		if frame:IsVisible() then
			frame:ClearAllPoints()
			if ( index == 1 ) then
				-- First bag
				frame:SetPoint("BOTTOMLEFT", frame:GetParent(), "BOTTOMRIGHT", -xOffset - maxwidth, yOffset );
				index = index + 1
				prevframe = frame
			elseif ( freeScreenHeight < frame:GetHeight() ) then
				-- Start a new column
				column = column + 1;
				freeScreenHeight = screenHeight - yOffset;
				frame:SetPoint("BOTTOMLEFT", frame:GetParent(), "BOTTOMRIGHT", -(column * maxwidth) - xOffset - maxwidth, yOffset );
				index = index + 1
				prevframe = frame
			else
				-- Anchor to the previous bag
				frame:SetPoint("BOTTOMLEFT", prevframe, "TOPLEFT", 0, CONTAINER_SPACING);	
				index = index + 1
				prevframe = frame
			end
			self:SaveBagPosition(id)
			freeScreenHeight = freeScreenHeight - frame:GetHeight() - VISIBLE_CONTAINER_SPACING;
		end
	end
end

function Baggins:OnTooltipUpdate()
	local cat = tablet:AddCategory() 
	
	for bagid, bag in ipairs(self.db.profile.bags) do
		cat:AddLine("text",bag.name,"func", function(self, bagid) self:ToggleBag(bagid) end ,"arg1",self,"arg2",bagid)
	end
	tablet:SetHint("Click a bag to toggle it")
end


function Baggins:OpenBagEditTablet()
	tablet:Open("BagginsEditBags")
end

function Baggins:OpenCategoryEditTablet()
	tablet:Open("BagginsEditCategories")
end
local catsorttable = {}
function Baggins:FillBagEditTablet()
	local cat
	tablet:SetTitle("Bags")
	tablet:SetTitleColor(1,1,1)
	for bagid, bag in ipairs(self.db.profile.bags) do
		if currentbag ~= bagid then
			cat = tablet:AddCategory("id",bagid,"text","+ "..bag.name,"size",14,"showWithoutChildren",true,"hideBlankLine",true,
										"func",function(bagid) 
											if IsControlKeyDown() then 
												self:StartDeleteBag(bagid) 
											else 
												currentbag = bagid
												currentsection = nil
											end 
											tablet:Refresh("BagginsEditBags") 
										end,
										"arg1",bagid)
		else
			cat = tablet:AddCategory("id",bagid,"text","- "..bag.name,"size",14,"showWithoutChildren",true,"hideBlankLine",true,
										"func",function(bagid) 
											if IsControlKeyDown() then 
												self:StartDeleteBag(bagid) 
											else 
												currentbag = nil 
												currentsection = nil
											end 
											tablet:Refresh("BagginsEditBags") 
										end,
										"arg1",bagid)
										
			cat:AddLine("text","Options","size",12,"indentation",10)
			cat:AddLine("text","Open With All","size",11,"indentation",15,"func",function(bag) bag.openWithAll = not bag.openWithAll tablet:Refresh("BagginsEditBags") end,"arg1",bag,"hasCheck",true,"checked",bag.openWithAll,"textR",1,"textG",1,"textB",1)
			cat:AddLine("text","Sections","size",12,"indentation",10)
			for sectionid, section in ipairs(bag.sections) do
				if sectionid ~= currentsection then
					cat:AddLine("text","+ "..section.name,"size",12,"indentation", 20,"textR",1,"textG",1,"textB",1,
									"func",function(bagid,sectionid) 
										if IsControlKeyDown() then 
											self:StartDeleteSection(bagid,sectionid) 
										else 
											currentsection = sectionid
										end 
										tablet:Refresh("BagginsEditBags") 
									end,
									"arg1",bagid,"arg2",sectionid)
				else
					cat:AddLine("text","- "..section.name,"size",12,"indentation", 20,"textR",1,"textG",1,"textB",1,
									"func",function(bagid,sectionid) 
										if IsControlKeyDown() then 
											self:StartDeleteSection(bagid,sectionid) 
										else 
											currentsection = nil
										end 
										tablet:Refresh("BagginsEditBags") 
									end,
									"arg1",bagid,"arg2",sectionid)
									
					cat:AddLine("text","Categories","size",10,"indentation",30)
					for categoryid, category in pairs(section.cats) do
						cat:AddLine("text",category,"size",10,"indentation",40,"textR",1,"textG",1,"textB",1,
							"func",function(bagid,sectionid) 
									if IsControlKeyDown() then
										self:StartDeleteCategory(bagid,sectionid,categoryid) 
									end
							end,
							"arg1",bagid,"arg2",sectionid)
					end		
					--cat:AddLine("text"," ","size",10)
					cat:AddLine("text","Add Category","size",10,"indentation", 40,"textR",0.2,"textG",1,"textB",0.5,
							"func",function(bagid,sectionid) 
									dewdrop:Open(self.dewdropparent, "children", function()
																						while #catsorttable > 0 do
																							tremove(catsorttable,#catsorttable)
																						end
																						for catid in pairs(self.db.profile.categories) do
																							tinsert(catsorttable,catid)
																						end
																						table.sort(catsorttable)

																						for k, catid in ipairs(catsorttable) do
																						dewdrop:AddLine("text",catid,"func",function(category) self:AddCategory(bagid,sectionid,category) tablet:Refresh("BagginsEditBags") end,"arg1",catid,"closeWhenClicked",true)
																					end
																				end,
																	'point', "TOPLEFT",
																	'relativePoint', "TOPLEFT",
																	'cursorX', true, "cursorY", true)
							end,
							"arg1",bagid,"arg2",sectionid)
					cat:AddLine("text"," ","size",10)
				end

			end
			--cat:AddLine("text"," ","size",12)
			cat:AddLine("text","New Section","size",12,"indentation", 20,"textR",0.2,"textG",1,"textB",0.5,
							"func",function(bagid,sectionid) 
									StaticPopup_Show("BAGGINS_NEW_SECTION") 
							end,
							"arg1",bagid,"arg2",sectionid)
		end
	end
	cat = tablet:AddCategory("text","Add New Bag","size",14,"textR",0.2,"textG",1,"textB",0.5,"showWithoutChildren",true,"func", function() StaticPopup_Show("BAGGINS_NEW_BAG") end)
	cat:AddLine("text"," ","size",12)
	--ugh calling Close directly here breaks things, ugly hack!
	cat:AddLine("text","Close","func", function() self:ScheduleEvent(function() tablet:Close("BagginsEditBags") end, 0.01) end)
	tablet:SetHint("Click on an entry to open it, Ctrl-Click to delete.")
end


function Baggins:FillCategoryEditTablet()
	local cat
	tablet:SetTitle("Categories")
	tablet:SetTitleColor(1,1,1)
	
	while #catsorttable > 0 do
		tremove(catsorttable,#catsorttable)
	end
	for catid in pairs(self.db.profile.categories) do
		tinsert(catsorttable,catid)
	end
	table.sort(catsorttable)
	
	for k, catid in ipairs(catsorttable) do
		local category = self.db.profile.categories[catid]
		if currentcategory ~= catid then
			cat = tablet:AddCategory("id",catid,"text","+ "..catid,"size",14,"showWithoutChildren",true,"hideBlankLine",true,
										"func",function(catid) 
											if IsControlKeyDown() then 
												self:StartDeleteCategory(catid) 
											else 
												currentcategory = catid
											end 
											tablet:Refresh("BagginsEditCategories") 
										end,
										"arg1",catid)
		else
			cat = tablet:AddCategory("id",catid,"text","- "..catid,"size",14,"showWithoutChildren",true,"hideBlankLine",true,
										"func",function(catid) 
											if IsControlKeyDown() then 
												self:StartDeleteCategory(catid) 
											else 
												currentcategory = nil
											end 
											tablet:Refresh("BagginsEditCategories") 
										end,
										"arg1",catid)
			cat:AddLine("text","Rules","size",12,"indentation",10)
			for ruleid, rule in ipairs(category) do
				local rulename
				if rule.type == "New" then
					rulename = "New Rule"
				else
					rulename = RuleTypes[rule.type].GetName(rule)
				end
				if ruleid > 1 then
					rulename = (rule.operation or "OR").." "..rulename
				end
				cat:AddLine("text",rulename,"size",10,"indentation",20,"textR",1,"textG",1,"textB",1,
					"func",function(catid,ruleid,rule) 
							if IsControlKeyDown() then
								self:StartDeleteRule(catid,ruleid) 
							else
								dewdrop:Open(self.dewdropparent, "children", function(...) Baggins:CreateRulesDewdrop(ruleid, rule, ...) end,
															'point', "TOPLEFT",
															'relativePoint', "TOPLEFT",
															'cursorX', true, "cursorY", true)
							end
					end,
					"arg1",catid,"arg2",ruleid,"arg3",rule)
			end		
			cat:AddLine("text","Add Rule","size",10,"indentation", 20,"textR",0.2,"textG",1,"textB",0.5,
					"func",function() 
							tinsert(category, { type="New" })
					end,
					"arg1",bagid,"arg2",sectionid)
			cat:AddLine("text"," ","size",10)
		end
	end
	cat = tablet:AddCategory("text","Add New Category","size",14,"textR",0.2,"textG",1,"textB",0.5,"showWithoutChildren",true,"func", function() StaticPopup_Show("BAGGINS_NEW_CATEGORY") end)
	cat:AddLine("text"," ","size",12)
	--ugh calling Close directly here breaks things, ugly hack!
	cat:AddLine("text","Close","func", function() self:ScheduleEvent(function() tablet:Close("BagginsEditCategories") end, 0.01) end)
	tablet:SetHint("Click on an entry to open it, Ctrl-Click to delete.")
end

function Baggins:CreateRulesDewdrop(ruleindex, rule, level, value, valueN_1, valueN_2, valueN_3, valueN_4)
	if level == 1 then
		dewdrop:AddLine('text', "Editing Rule", 'isTitle', true)
		
		dewdrop:AddLine('text', "Type", 'hasArrow', true,'value',"Type")
		dewdrop:AddLine()
		if not rule.type or rule.type == "New" then
			dewdrop:AddLine('text', "Select a rule type to create the rule", "isTitle",true)
		else
			RuleTypes[rule.type].DewDropOptions(rule, level, value)
			if ruleindex > 1 then
				dewdrop:AddLine('text', "Operation", 'hasArrow', true,'value',"Operation")
			end
		end
	elseif level == 2 and value == "Type" then
		for ruletype, ruledef in pairs(RuleTypes) do
			dewdrop:AddLine('text', ruledef.DisplayName, "checked", rule.type == rulename,
				"tooltipTitle", ruledef.DisplayName, "tooltipText", ruledef.Description,
				"func",function(ruletype, rule) rule.type = ruletype self:CleanRule(rule) self:OnRuleChanged() end, "arg1", ruletype, "arg2", rule)
		end
	elseif level == 2 and value == "Operation" then
		dewdrop:AddLine('text', "AND", "checked", rule.operation == "AND","func",function() rule.operation = "AND" self:OnRuleChanged() end)
		dewdrop:AddLine('text', "OR", "checked", rule.operation == "OR" or rule.operation == nil,"func",function() rule.operation = "OR" self:OnRuleChanged() end)
		dewdrop:AddLine('text', "NOT", "checked", rule.operation == "NOT","func",function() rule.operation = "NOT" self:OnRuleChanged() end)
	elseif level == 2 then
		RuleTypes[rule.type].DewDropOptions(rule,level,value)
	end
		
end

function Baggins:OnRuleChanged()
	self:ForceFullUpdate()
	self:UpdateBags()
	tablet:Refresh("BagginsEditCategories")
end

local deletingbag
local deletingsection
local deletingcategory
local deletingrule
function Baggins:StartDeleteBag(bagid)
	deletingbag = bagid
	StaticPopup_Show("BAGGINS_DEL_BAG")
end

function Baggins:ConfirmDeleteBag()
	self:RemoveBag(deletingbag)
	deletingbag = nil
end

function Baggins:StartDeleteSection(bagid,sectionid)
	deletingbag = bagid
	deletingsection = sectionid
	StaticPopup_Show("BAGGINS_DEL_SECTION")
end

function Baggins:ConfirmDeleteSection()
	self:RemoveSection(deletingbag,deletingsection)
	deletingbag = nil
	deletingsection = nil
end

function Baggins:StartDeleteCategory(bagid,sectionid,categoryid)
	deletingbag = bagid
	deletingsection = sectionid
	deletingcategory = categoryid
	StaticPopup_Show("BAGGINS_DEL_CATEGORY")
end

function Baggins:ConfirmDeleteCategory()
	self:RemoveCategory(deletingbag,deletingsection,deletingcategory)
	deletingbag = nil
	deletingsection = nil
	deletingcategory = nil
end

function Baggins:StartDeleteRule(catid,ruleid)
	deletingcat = catid
	deletingrule = ruleid
	StaticPopup_Show("BAGGINS_DEL_RULE")
end

function Baggins:ConfirmDeleteRule()
	self:RemoveRule(deletingcat,deletingrule)
	Baggins:OnRuleChanged()
	deletingcat = nil
	deletingrule = nil
end

StaticPopupDialogs["BAGGINS_NEW_BAG"] = {
	text = "Baggins - New Bag",
	button1 = "Accept",
	button2 = "Cancel",
	hasEditBox = 1,
	maxLetters = 20,
	OnAccept = function()
		local editBox = getglobal(this:GetParent():GetName().."EditBox");
		Baggins:NewBag(editBox:GetText());
		tablet:Refresh("BagginsEditBags")
	end,
	OnShow = function()
		getglobal(this:GetName().."EditBox"):SetFocus();
	end,
	OnHide = function()
		if ( ChatFrameEditBox:IsVisible() ) then
			ChatFrameEditBox:SetFocus();
		end
		getglobal(this:GetName().."EditBox"):SetText("");
	end,
	EditBoxOnEnterPressed = function()
		local editBox = getglobal(this:GetParent():GetName().."EditBox");
		Baggins:NewBag(editBox:GetText());
		tablet:Refresh("BagginsEditBags")
		this:GetParent():Hide();
	end,
	EditBoxOnEscapePressed = function()
		this:GetParent():Hide();
	end,
	timeout = 0,
	exclusive = 1,
	whileDead = 1,
	hideOnEscape = 1
}
StaticPopupDialogs["BAGGINS_NEW_SECTION"] = {
	text = "Baggins - New Section",
	button1 = "Accept",
	button2 = "Cancel",
	hasEditBox = 1,
	maxLetters = 20,
	OnAccept = function()
		local editBox = getglobal(this:GetParent():GetName().."EditBox");
		Baggins:NewSection(currentbag,editBox:GetText());
		tablet:Refresh("BagginsEditBags")
	end,
	OnShow = function()
		getglobal(this:GetName().."EditBox"):SetFocus();
	end,
	OnHide = function()
		if ( ChatFrameEditBox:IsVisible() ) then
			ChatFrameEditBox:SetFocus();
		end
		getglobal(this:GetName().."EditBox"):SetText("");
	end,
	EditBoxOnEnterPressed = function()
		local editBox = getglobal(this:GetParent():GetName().."EditBox");
		Baggins:NewSection(currentbag,editBox:GetText());
		tablet:Refresh("BagginsEditBags")
		this:GetParent():Hide();
	end,
	EditBoxOnEscapePressed = function()
		this:GetParent():Hide();
	end,
	timeout = 0,
	exclusive = 1,
	whileDead = 1,
	hideOnEscape = 1
}
StaticPopupDialogs["BAGGINS_NEW_CATEGORY"] = {
	text = "Baggins - New Category",
	button1 = "Accept",
	button2 = "Cancel",
	hasEditBox = 1,
	maxLetters = 20,
	OnAccept = function()
		local editBox = getglobal(this:GetParent():GetName().."EditBox");
		Baggins:NewCategory(editBox:GetText());
		tablet:Refresh("BagginsEditCategories")
	end,
	OnShow = function()
		getglobal(this:GetName().."EditBox"):SetFocus();
	end,
	OnHide = function()
		if ( ChatFrameEditBox:IsVisible() ) then
			ChatFrameEditBox:SetFocus();
		end
		getglobal(this:GetName().."EditBox"):SetText("");
	end,
	EditBoxOnEnterPressed = function()
		local editBox = getglobal(this:GetParent():GetName().."EditBox");
		local name = editBox:GetText()
		this:GetParent():Hide();
		Baggins:NewCategory(name);
		tablet:Refresh("BagginsEditCategories")
		
	end,
	EditBoxOnEscapePressed = function()
		this:GetParent():Hide();
	end,
	timeout = 0,
	exclusive = 1,
	whileDead = 1,
	hideOnEscape = 1
}
StaticPopupDialogs["BAGGINS_DEL_BAG"] = {
	text = "Are you sure you want to delete this bag? this cannot be undone",
	button1 = "Delete",
	button2 = "Cancel",
	OnAccept = function()
		Baggins:ConfirmDeleteBag();
		tablet:Refresh("BagginsEditBags")
	end,
	timeout = 0,
	exclusive = 1,
	hideOnEscape = 1
}
StaticPopupDialogs["BAGGINS_DEL_SECTION"] = {
	text = "Are you sure you want to delete this Section? this cannot be undone",
	button1 = "Delete",
	button2 = "Cancel",
	OnAccept = function()
		Baggins:ConfirmDeleteSection();
		tablet:Refresh("BagginsEditBags")
	end,
	timeout = 0,
	exclusive = 1,
	hideOnEscape = 1
}
StaticPopupDialogs["BAGGINS_DEL_CATEGORY"] = {
	text = "Are you sure you want to remove this Category? this cannot be undone",
	button1 = "Delete",
	button2 = "Cancel",
	OnAccept = function()
		Baggins:ConfirmDeleteCategory();
		tablet:Refresh("BagginsEditBags")
		tablet:Refresh("BagginsEditCategories")
	end,
	timeout = 0,
	exclusive = 1,
	hideOnEscape = 1
}
StaticPopupDialogs["BAGGINS_DEL_RULE"] = {
	text = "Are you sure you want to remove this Rule? this cannot be undone",
	button1 = "Delete",
	button2 = "Cancel",
	OnAccept = function()
		Baggins:ConfirmDeleteRule();
		tablet:Refresh("BagginsEditCategories")
	end,
	timeout = 0,
	exclusive = 1,
	hideOnEscape = 1
}
StaticPopupDialogs["BAGGINS_CATINUSE"] = {
	text = "That category is in use by one or more bags, you cannot delete it.",
	button1 = "Ok",
	timeout = 0,
	exclusive = 1,
	hideOnEscape = 1
}
StaticPopupDialogs["BAGGINS_CATEXISTS"] = {
	text = "A category with that name already exists.",
	button1 = "Ok",
	timeout = 0,
	exclusive = 1,
	hideOnEscape = 1
}


function Baggins:RegisterEditTablets()
	tablet:Register("BagginsEditBags" ,
					"children", function() self:FillBagEditTablet() end,
					"data", {},
					"cantAttach", true,
					"clickable", true,
					"point", function() return "TOPLEFT", "TOPLEFT" end,
					"parent", "UIParent",
					"showTitleWhenDetached", true,
					"showHintWhenDetached", true,
					"dontHook", true,
					"strata","DIALOG",
					"framelevel",20
					)				
	tablet:Register("BagginsEditCategories" ,
					"children", function() self:FillCategoryEditTablet() end,
					"data", {},
					"cantAttach", true,
					"clickable", true,
					"point", function() return "TOPLEFT", "TOPLEFT" end,
					"parent", "UIParent",
					"showTitleWhenDetached", true,
					"showHintWhenDetached", true,
					"dontHook", true,
					"strata","DIALOG"
					)		
	self.dewdropparent = CreateFrame("frame",nil,UIParent)
	self.dewdropparent:SetHeight(1)
	self.dewdropparent:SetWidth(1)
	self.dewdropparent:SetPoint("CENTER",UIParent,"CENTER",0,0)
end

